@@ -2,6 +2,7 @@
 
 import sys
 import types
+from types import LambdaType
 from collections import defaultdict
 from collections.abc import Iterable
 from itertools import product
@@ -16,6 +17,7 @@ from openmdao.core.system import System, _supported_methods, _DEFAULT_COLORING_M
 from openmdao.core.constants import INT_DTYPE
 from openmdao.jacobians.dictionary_jacobian import DictionaryJacobian
 from openmdao.utils.array_utils import shape_to_len
+from openmdao.utils.class_util import overrides_method
 from openmdao.utils.units import simplify_unit
 from openmdao.utils.name_maps import abs_key_iter, abs_key2rel_key, rel_name2abs_name
 from openmdao.utils.mpi import MPI
@@ -25,6 +27,8 @@ from openmdao.utils.indexer import Indexer, indexer
 import openmdao.utils.coloring as coloring_mod
 from openmdao.utils.om_warnings import issue_warning, MPIWarning, DistributedComponentWarning, \
     DerivativesWarning, warn_deprecation
+from openmdao.utils.code_utils import is_lambda, LambdaPickleWrapper
+
 
 _forbidden_chars = {'.', '*', '?', '!', '[', ']'}
 _whitespace = {' ', '\t', '\r', '\n'}
@@ -306,9 +310,49 @@ class Component(System):
         else:
             self._discrete_inputs = self._discrete_outputs = ()
 
+        if self.comm.size > 1:
+            # check that same variables are declared on all procs
+            vnames = (list(self._var_rel_names['output']), list(self._var_rel_names['input']))
+            allnames = self.comm.gather(vnames, root=0)
+            if self.comm.rank == 0:
+                outset, inset = vnames
+                msg = ''
+                for oset, iset in allnames:
+                    if iset != inset or oset != outset:
+                        msg = self._missing_vars_error(allnames)
+                        break
+                self.comm.bcast(msg, root=0)
+            else:
+                msg = self.comm.bcast(None, root=0)
+
+            if msg:
+                raise RuntimeError(msg)
+
         self._serial_idxs = None
         self._inconsistent_keys = set()
 
+    def _missing_vars_error(self, allnames):
+        msg = ''
+        outset, inset = allnames[0]
+        for rank, (olist, ilist) in enumerate(allnames):
+            if rank != 0 and (olist != outset or ilist != inset):
+                idiff = set(inset).symmetric_difference(ilist)
+                odiff = set(outset).symmetric_difference(olist)
+                if idiff or odiff:
+                    varnames = sorted(idiff | odiff)
+                    if len(varnames) == 1:
+                        varmsg = f"Variable '{varnames[0]}' exists on some ranks and not others."
+                    else:
+                        varmsg = f"Variables {varnames} exist on some ranks and not others."
+                else:
+                    varmsg = "Variables have not been declared in the same order on all ranks."
+
+                msg = (f"{self.msginfo}: {varmsg} A component must declare all variables in "
+                       "the same order on all ranks, even if the size of the variable is 0 on "
+                       "some ranks.")
+                break
+        return msg
+
     @collect_errors
     def _setup_var_sizes(self):
         """
@@ -366,6 +410,12 @@ class Component(System):
         """
         pass
 
+    def _setup_residuals(self):
+        """
+        Process hook to call user-defined setup_residuals method if provided.
+        """
+        pass
+
     def _declared_partials_iter(self):
         """
         Iterate over all declared partials.
@@ -441,8 +491,7 @@ class Component(System):
         return False
 
     def _promoted_wrt_iter(self):
-        _, wrts = self._get_partials_varlists()
-        yield from wrts
+        yield from self._get_partials_wrts()
 
     def _update_subjac_sparsity(self, sparsity):
         """
@@ -557,11 +606,13 @@ class Component(System):
             if distributed is None:
                 distributed = False
             # using ._dict below to avoid tons of deprecation warnings
-            distributed = distributed or self.options._dict['distributed']['val']
+            distributed = distributed or ('distributed' in self.options and
+                                          self.options._dict['distributed']['val'])
 
-        metadata = {}
+        if compute_shape is not None and is_lambda(compute_shape):
+            compute_shape = LambdaPickleWrapper(compute_shape)
 
-        metadata.update({
+        metadata = {
             'val': val,
             'shape': shape,
             'size': shape_to_len(shape),
@@ -574,7 +625,7 @@ class Component(System):
             'shape_by_conn': shape_by_conn,
             'compute_shape': compute_shape,
             'copy_shape': copy_shape,
-        })
+        }
 
         # this will get reset later if comm size is 1
         self._has_distrib_vars |= metadata['distributed']
@@ -798,7 +849,8 @@ class Component(System):
             if distributed is None:
                 distributed = False
             # using ._dict below to avoid tons of deprecation warnings
-            distributed = distributed or self.options._dict['distributed']['val']
+            distributed = distributed or ('distributed' in self.options and
+                                          self.options._dict['distributed']['val'])
 
         if copy_shape and compute_shape:
             raise ValueError(f"{self.msginfo}: Only one of 'copy_shape' or 'compute_shape' can "
@@ -812,9 +864,10 @@ class Component(System):
             raise TypeError(f"{self.msginfo}: The compute_shape argument should be a function but "
                             f"a '{type(compute_shape).__name__}' was given.")
 
-        metadata = {}
+        if compute_shape is not None and is_lambda(compute_shape):
+            compute_shape = LambdaPickleWrapper(compute_shape)
 
-        metadata.update({
+        metadata = {
             'val': val,
             'shape': shape,
             'size': shape_to_len(shape),
@@ -831,7 +884,7 @@ class Component(System):
             'shape_by_conn': shape_by_conn,
             'compute_shape': compute_shape,
             'copy_shape': copy_shape,
-        })
+        }
 
         # this will get reset later if comm size is 1
         self._has_distrib_vars |= metadata['distributed']
@@ -1018,25 +1071,14 @@ class Component(System):
         **kwargs : dict
             Keyword arguments for controlling the behavior of the approximation.
         """
-        pattern_matches = self._find_partial_matches(of, wrt)
         self._has_approx = True
+        info = self._subjacs_info
 
-        for of_bundle, wrt_bundle in product(*pattern_matches):
-            of_pattern, of_matches = of_bundle
-            wrt_pattern, wrt_matches = wrt_bundle
-            if not of_matches:
-                raise ValueError('{}: No matches were found for of="{}"'.format(self.msginfo,
-                                                                                of_pattern))
-            if not wrt_matches:
-                raise ValueError('{}: No matches were found for wrt="{}"'.format(self.msginfo,
-                                                                                 wrt_pattern))
-
-            info = self._subjacs_info
-            for abs_key in abs_key_iter(self, of_matches, wrt_matches):
-                meta = info[abs_key]
-                meta['method'] = method
-                meta.update(kwargs)
-                info[abs_key] = meta
+        for abs_key in self._matching_key_iter(of, wrt):
+            meta = info[abs_key]
+            meta['method'] = method
+            meta.update(kwargs)
+            info[abs_key] = meta
 
     def declare_partials(self, of, wrt, dependent=True, rows=None, cols=None, val=None,
                          method='exact', step=None, form=None, step_calc=None, minimum_step=None):
@@ -1131,8 +1173,8 @@ class Component(System):
                                  f'will raise an error.')
 
             if rows is not None:
-                rows = np.array(rows, dtype=INT_DTYPE, copy=False)
-                cols = np.array(cols, dtype=INT_DTYPE, copy=False)
+                rows = np.asarray(rows, dtype=INT_DTYPE)
+                cols = np.asarray(cols, dtype=INT_DTYPE)
 
                 # Check the length of rows and cols to catch this easy mistake and give a
                 # clear message.
@@ -1340,7 +1382,7 @@ class Component(System):
         if not self._declared_partial_checks:
             return {}
         opts = {}
-        _, wrt = self._get_partials_varlists()
+        wrt = self._get_partials_wrts()
         invalid_wrt = []
         matrix_free = self.matrix_free
 
@@ -1451,7 +1493,6 @@ class Component(System):
                 rows = None
                 cols = None
 
-        pattern_matches = self._find_partial_matches(of, '*' if wrt is None else wrt)
         abs2meta_in = self._var_abs2meta['input']
         abs2meta_out = self._var_abs2meta['output']
 
@@ -1459,7 +1500,134 @@ class Component(System):
         patmeta = dict(pattern_meta)
         patmeta_not_none = {k: v for k, v in pattern_meta.items() if v is not None}
 
-        for of_bundle, wrt_bundle in product(*pattern_matches):
+        for abs_key in self._matching_key_iter(of, '*' if wrt is None else wrt):
+            if not dependent:
+                if abs_key in self._subjacs_info:
+                    del self._subjacs_info[abs_key]
+                continue
+
+            if abs_key in self._subjacs_info:
+                meta = self._subjacs_info[abs_key]
+                meta.update(patmeta_not_none)
+            else:
+                meta = patmeta.copy()
+
+            of, wrt = abs_key
+            meta['rows'] = rows
+            meta['cols'] = cols
+            csz = abs2meta_in[wrt]['size'] if wrt in abs2meta_in else abs2meta_out[wrt]['size']
+            meta['shape'] = shape = (abs2meta_out[of]['size'], csz)
+            dist_out = abs2meta_out[of]['distributed']
+            if wrt in abs2meta_in:
+                dist_in = abs2meta_in[wrt]['distributed']
+            else:
+                dist_in = abs2meta_out[wrt]['distributed']
+
+            if dist_in and not dist_out and not self.matrix_free:
+                rel_key = abs_key2rel_key(self, abs_key)
+                raise RuntimeError(f"{self.msginfo}: component has defined partial {rel_key} "
+                                   "which is a non-distributed output wrt a distributed input."
+                                   " This is only supported using the matrix free API.")
+
+            if shape[0] == 0 or shape[1] == 0:
+                msg = "{}: '{}' is an array of size 0"
+                if shape[0] == 0:
+                    if dist_out:
+                        # distributed vars are allowed to have zero size inputs on some procs
+                        rows_max = -1
+                    else:
+                        # non-distributed vars are not allowed to have zero size inputs
+                        raise ValueError(msg.format(self.msginfo, of))
+                if shape[1] == 0:
+                    if not dist_in:
+                        # non-distributed vars are not allowed to have zero size outputs
+                        raise ValueError(msg.format(self.msginfo, wrt))
+                    else:
+                        # distributed vars are allowed to have zero size outputs on some procs
+                        cols_max = -1
+
+            if val is None and not matfree:
+                # we can only get here if rows is None  (we're not sparse list format)
+                meta['val'] = np.zeros(shape)
+            elif is_array:
+                if rows is None and val.shape != shape and val.size == shape[0] * shape[1]:
+                    meta['val'] = val = val.copy().reshape(shape)
+                else:
+                    meta['val'] = val.copy()
+            elif is_scalar:
+                meta['val'] = np.full(shape, val, dtype=float)
+            else:
+                meta['val'] = val
+
+            if rows_max >= shape[0] or cols_max >= shape[1]:
+                of, wrt = abs_key2rel_key(self, abs_key)
+                raise ValueError(f"{self.msginfo}: d({of})/d({wrt}): Expected {shape[0]}x"
+                                 f"{shape[1]} but declared at least {rows_max + 1}x"
+                                 f"{cols_max + 1}")
+
+            self._check_partials_meta(abs_key, meta['val'],
+                                      shape if rows is None else (rows.shape[0], 1))
+
+            self._subjacs_info[abs_key] = meta
+
+    def _get_partials_wrts(self):
+        """
+        Get list of 'wrt' variables that form the partial jacobian.
+
+        Returns
+        -------
+        list
+            List of 'wrt' relative variable names.
+        """
+        # filter out any discrete inputs or outputs
+        if self._discrete_inputs:
+            return [n for n in self._var_rel_names['input'] if n not in self._discrete_inputs]
+
+        return list(self._var_rel_names['input'])
+
+    def _get_partials_ofs(self, use_resname=False):
+        """
+        Get lists of 'of' variables that form the partial jacobian.
+
+        Parameters
+        ----------
+        use_resname : bool
+            Ignored.
+
+        Returns
+        -------
+        list
+            List of 'of' relative variable names.
+        """
+        # filter out any discrete inputs or outputs
+        if self._discrete_outputs:
+            return [n for n in self._var_rel_names['output'] if n not in self._discrete_outputs]
+
+        return list(self._var_rel_names['output'])
+
+    def _matching_key_iter(self, of_patterns, wrt_patterns, use_resname=False):
+        """
+        Iterate over all combinations of matching keys for the given patterns.
+
+        Parameters
+        ----------
+        of_patterns : list of str
+            List of variable names and/or glob patterns for the 'of' variables.
+        wrt_patterns : list of str
+            List of variable names and/or glob patterns for the 'wrt' variables.
+        use_resname : bool, optional
+            If True, match of_patterns against residuals instead of outputs.
+
+        Yields
+        ------
+        tuple
+            A tuple of matching keys, where the first element is the 'of' key and the second
+            element is the 'wrt' key.  Both are absolute names.
+        """
+        of_bundles = self._find_of_matches(of_patterns, use_resname=use_resname)
+        wrt_bundles = self._find_wrt_matches(wrt_patterns)
+
+        for of_bundle, wrt_bundle in product(of_bundles, wrt_bundles):
             of_pattern, of_matches = of_bundle
             wrt_pattern, wrt_matches = wrt_bundle
             if not of_matches:
@@ -1468,107 +1636,46 @@ class Component(System):
             if not wrt_matches:
                 raise ValueError('{}: No matches were found for wrt="{}"'.format(self.msginfo,
                                                                                  wrt_pattern))
+            yield from abs_key_iter(self, of_matches, wrt_matches)
 
-            for abs_key in abs_key_iter(self, of_matches, wrt_matches):
-                if not dependent:
-                    if abs_key in self._subjacs_info:
-                        del self._subjacs_info[abs_key]
-                    continue
-
-                if abs_key in self._subjacs_info:
-                    meta = self._subjacs_info[abs_key]
-                    meta.update(patmeta_not_none)
-                else:
-                    meta = patmeta.copy()
-
-                of, wrt = abs_key
-                meta['rows'] = rows
-                meta['cols'] = cols
-                csz = abs2meta_in[wrt]['size'] if wrt in abs2meta_in else abs2meta_out[wrt]['size']
-                meta['shape'] = shape = (abs2meta_out[of]['size'], csz)
-                dist_out = abs2meta_out[of]['distributed']
-                if wrt in abs2meta_in:
-                    dist_in = abs2meta_in[wrt]['distributed']
-                else:
-                    dist_in = abs2meta_out[wrt]['distributed']
-
-                if dist_in and not dist_out and not self.matrix_free:
-                    rel_key = abs_key2rel_key(self, abs_key)
-                    raise RuntimeError(f"{self.msginfo}: component has defined partial {rel_key} "
-                                       "which is a non-distributed output wrt a distributed input."
-                                       " This is only supported using the matrix free API.")
-
-                if shape[0] == 0 or shape[1] == 0:
-                    msg = "{}: '{}' is an array of size 0"
-                    if shape[0] == 0:
-                        if dist_out:
-                            # distributed vars are allowed to have zero size inputs on some procs
-                            rows_max = -1
-                        else:
-                            # non-distributed vars are not allowed to have zero size inputs
-                            raise ValueError(msg.format(self.msginfo, of))
-                    if shape[1] == 0:
-                        if not dist_in:
-                            # non-distributed vars are not allowed to have zero size outputs
-                            raise ValueError(msg.format(self.msginfo, wrt))
-                        else:
-                            # distributed vars are allowed to have zero size outputs on some procs
-                            cols_max = -1
-
-                if val is None and not matfree:
-                    # we can only get here if rows is None  (we're not sparse list format)
-                    meta['val'] = np.zeros(shape)
-                elif is_array:
-                    if rows is None and val.shape != shape and val.size == shape[0] * shape[1]:
-                        meta['val'] = val = val.copy().reshape(shape)
-                    else:
-                        meta['val'] = val.copy()
-                elif is_scalar:
-                    meta['val'] = np.full(shape, val, dtype=float)
-                else:
-                    meta['val'] = val
-
-                if rows_max >= shape[0] or cols_max >= shape[1]:
-                    of, wrt = abs_key2rel_key(self, abs_key)
-                    raise ValueError(f"{self.msginfo}: d({of})/d({wrt}): Expected {shape[0]}x"
-                                     f"{shape[1]} but declared at least {rows_max + 1}x"
-                                     f"{cols_max + 1}")
-
-                self._check_partials_meta(abs_key, meta['val'],
-                                          shape if rows is None else (rows.shape[0], 1))
-
-                self._subjacs_info[abs_key] = meta
-
-    def _find_partial_matches(self, of_pattern, wrt_pattern, use_resname=False):
+    def _find_of_matches(self, pattern, use_resname=False):
         """
-        Find all partial derivative matches from of_pattern and wrt_pattern.
+        Find all matches for the given 'of' pattern.
 
         Parameters
         ----------
-        of_pattern : str or list of str
-            The relative name(s) of the residual(s) that derivatives are being computed for.
-            May also contain glob patterns.
-        wrt_pattern : str or list of str
-            The relative name(s) of the variable(s) that derivatives are taken with respect to.
-            Each name can refer to an input or an output variable. May also contain glob patterns.
+        pattern : str
+            Glob pattern or relative variable name.
         use_resname : bool
-            If True, use residual names for 'of' patterns.
+            If True, match residual names instead of output names.
 
         Returns
         -------
-        tuple(list, list)
-            Pair of lists containing pattern matches (if any). Returns (of_matches, wrt_matches)
-            where of_matches is a list of tuples (pattern, matches) and wrt_matches is a list of
-            tuples (pattern, output_matches, input_matches).
+        list
+            List of tuples of the form (abs_name, meta) where abs_name is the absolute name of the
+            matching variable and meta is the metadata for that variable.
         """
-        ofs, wrts = self._get_partials_varlists(use_resname=use_resname)
+        of_list = [pattern] if isinstance(pattern, str) else pattern
+        return [(pattern, find_matches(pattern, self._get_partials_ofs(use_resname=use_resname)))
+                for pattern in of_list]
 
-        of_list = [of_pattern] if isinstance(of_pattern, str) else of_pattern
-        wrt_list = [wrt_pattern] if isinstance(wrt_pattern, str) else wrt_pattern
+    def _find_wrt_matches(self, pattern):
+        """
+        Find all matches for the given 'wrt' pattern.
 
-        of_pattern_matches = [(pattern, find_matches(pattern, ofs)) for pattern in of_list]
-        wrt_pattern_matches = [(pattern, find_matches(pattern, wrts)) for pattern in wrt_list]
-        return of_pattern_matches, wrt_pattern_matches
+        Parameters
+        ----------
+        pattern : str
+            Glob pattern or relative variable name.
+
+        Returns
+        -------
+        list
+            List of tuples of the form (abs_name, meta) where abs_name is the absolute name of the
+            matching variable and meta is the metadata for that variable.
+        """
+        wrt_list = [pattern] if isinstance(pattern, str) else pattern
+        return [(pattern, find_matches(pattern, self._get_partials_wrts())) for pattern in wrt_list]
 
     def _check_partials_meta(self, abs_key, val, shape):
         """
@@ -1766,6 +1873,19 @@ class Component(System):
         """
         return True
 
+    def _get_graph_node_meta(self):
+        """
+        Return metadata to add to this system's graph node.
+
+        Returns
+        -------
+        dict
+            Metadata for this system's graph node.
+        """
+        meta = super()._get_graph_node_meta()
+        meta['base'] = 'ExplicitComponent' if self.is_explicit() else 'ImplicitComponent'
+        return meta
+
 
 class _DictValues(object):
     """
